#ifndef MUSCLES_TABU_SEARCH_H
#define MUSCLES_TABU_SEARCH_H

#include <algorithm>
#include <list>
#include <random>

template<typename Solution, typename RandomEngine>
Solution mutateSwap(Solution const&initial, RandomEngine &gen) {
	Solution solution = initial;
	std::size_t const&size = solution.size();

	if(size > 1) {
		std::uniform_int_distribution<std::size_t> dist{0, size-1};

		std::size_t i = dist(gen), j;
		do j = dist(gen); while(i == j);

		solution.swap(i, j);
	}

	return solution;
}

template<typename Solution, typename TabuList>
Solution bestNeighbour(Solution const&solution, TabuList& lTabu) {
	bool found = false;

	Solution best;

	for(std::size_t i = 0; i < solution.size()-1; ++i) {
		for(std::size_t j = i+1; j < solution.size(); ++j) {
			Solution neighbour{solution};
			neighbour.swap(i, j);

			typename std::decay_t<decltype(lTabu)>::const_iterator it;
			it = std::find(std::begin(lTabu), std::end(lTabu), neighbour);
			if(it == std::end(lTabu)) {
				if(neighbour < best) {
					best = neighbour;
					found = true;
				}
			}
		}
	}

	if(found) lTabu.push_back(best);

	return best;
}

template<typename Solution, typename TabuList>
Solution bestNeighbour2Opt(Solution const&solution, TabuList& lTabu) {
	bool found = false;

	Solution best;
	if(solution.size() < 3) return best;

	for(std::size_t i = 1; i < solution.size()-1; ++i) {
		for(std::size_t j = i+1; j < solution.size(); ++j) {
			if(i == 1 && j == solution.size()-1) continue;
			if(solution.value2Opt(i, j) < best.value()) {
				Solution neighbour{solution};
				neighbour.move2Opt(i, j);

				typename std::decay_t<decltype(lTabu)>::const_iterator it;
				it = std::find(std::begin(lTabu), std::end(lTabu), neighbour);
				if(it == std::end(lTabu)) {
					if(neighbour < best) {
						best = neighbour;
						found = true;
					}
				}
			}
		}
	}

	if(found) lTabu.push_back(best);

	return best;
}

template<typename Solution, typename Problem, typename Init, typename Neighbours>
Solution tabuSearch(Problem const&p, Init&& init, Neighbours&& findBestNeighbours) {
	std::mt19937 gen{std::random_device{}()};
	// std::mt19937 gen{0};
	Solution current, best;
	bool stop = false;
	std::list<Solution> lTabu;

	int notImproved = 0;

	best = current = init(p, gen);
	while(!stop) {
		current = findBestNeighbours(current, lTabu);
		if(current.empty()) stop = true;
		else {
			if(current < best) {
				best = current;
				notImproved = 0;
			} else {
				++notImproved;
				if(notImproved > 200) stop = true;
			}
		}
	}

	/*
	if(lTabu.size() > 1) {
		std::size_t n{};
		for(auto i = 0; i < lTabu.size()-1; ++i) {
			for(auto j = i+1; j < lTabu.size(); ++j) {
				std::size_t offset{};
				auto const&a = *std::next(std::begin(lTabu), i);
				auto const&b = *std::next(std::begin(lTabu), j);
				for(offset = 0; offset < a.size(); ++offset)
					if(a[offset] == b[0]) break;

				bool id = true;
				for(auto k = 0; id && k < a.size(); ++k)
					id &= a[(k+offset)%a.size()] == b[k];
				if(id) ++n;
			}
		}
	}
	*/

	return best;
}

#endif
